#!/bin/sh
#================================================================================
# (C)2013-2018 dMajo
# Title        : ad-blocker.sh
# Version      : V2.03.0111
# Author       : dMajo (http://forum.synology.com/enu/memberlist.php?mode=viewprofile&u=69661)
# Description  : Script to block add-banner servers, dns based
# Dependencies : Syno DNSServer package, bash V4 or higher, sed, sort, wget
# Usage        : sh ad-blocker.sh
#--------------------------------------------------------------------------------
# Version history:
# 2013.09.01 - 1.00.0001: Initial release
# 2013.09.08 - 1.00.0004: Fix: changed include target to support views
# 2013.09.12 - 1.00.0005: Added automatic zone file generation and some basic error handling
# 2014.03.29 - 1.01.0013: Added dependencies check
# 2014.03.30 - 1.02.0017: Script reorganized
# 2014.04.06 - 1.02.0018: Fix: fixed serial number in zone file generation
# 2015.11.01 - 1.03.0031: Added additional blacklist sources (users request)
# 2015.11.02 - 1.04.0035: Added blacklist&whitelist filo option
# 2015.12.26 - 1.06.0041: Added check for existing dns zones (thanks Briolet)
# 2016.04.24 - 2.00.0051: Script refactoring with external config file
# 2016.04.25 - 2.00.0053: Change: config file structure
# 2016.04.30 - 2.01.0061: Fix: config file security vulnerability and minor bugs
# 2016.05.01 - 2.01.0072: Script reorganization and cleanup
# 2017.12.16 - 2.02.0086: Compatibility for new DSNServer version
# 2017.12.24 - 2.03.0101: Improved zone SN handling and fixed logging feature after Synology switch to DB
# 2018.01.07 - 2.03.0102: Fixed small bug in Type=2 data polishing routine (thanks M4RC)
# 2018.01.27 - 2.03.0103: Improved ad-blocker.db's header comments
# 2018.01.27 - 2.03.0105: Fix: improved capitals handling and fixed a minor copy&paste bug (thanks Briolet)
# 2018.01.28 - 2.03.0111: Improved logging and added IPv6 support
# 2018.01.29 - 2.03.0112: Fixed a small DB logging glitch (thanks Briolet)
# 2018.02.04 - 2.03.0115: Enhanced cleaning filter rules
# 2018.02.04 - 2.03.0118: Added User BL & WL comments handling
# xxxx.xx.xx - xxxxxxxxx: Briolet: changed some paths for private use
#================================================================================
# Ad-Blocker Config file
#--------------------------------------------------------------------------------
# User whitelist and blacklist path/file association.
# ---------------------------------------------------
# User's black/white lists are supported as a FQDN domains/hosts list, one per line, same as downloads of BL_Type=3
# USER_BL = /path/to/the/filename
# USER_WL = /path/to/the/filename
#--------------------------------------------------------------------------------
# Blacklist array definition
# --------------------------
# Blacklist configuration fields, indexing starts from 1
# BL_Items		= Number of total configured blacklist downloads (n=1..Items)
#
# BL_Name[n]	= Blacklist's friendly name
# BL_URL[n]		= URL address and parameters to retrive the ASCII BL test file
# BL_Type[n]	= 0 Synology zone file (no processing, taken as is)
#				  1 MSDNS/Bind standard zone file
#				  2 Standard hosts file
#				  3 FQDN domain/host(only) list
# BL_Zipped[n]	= 0/1 Indicates that download is single zipped file
# BL_Enabled[n]	= 0/1 If not enabled will not be processed
#================================================================================

# Define dirs, files and functions
ScriptDir="$(cd "$(dirname "${BASH_SOURCE[0]}")" && pwd)" # Path to the scriptfolder
ScriptCfg="${ScriptDir}/ad-blocker.conf"
ScriptLog="${ScriptDir}/ad-blocker.log"
TmpDir="${ScriptDir}/ad-blocker.tmp"
PackageDir="/var/packages/DNSServer"
RootDir="${PackageDir}/target"
LogDir="${RootDir}/named/var/log"
LogDB="${LogDir}/SYNODNSLOGDB"
LogFile="${LogDir}/default-log"
ZoneDir="${RootDir}/named/etc/zone"
ZoneDataDir="${ZoneDir}/data"
ZoneMasterDir="${ZoneDir}/master"
Debug=1 # If debug=1 the temp folder is not deleted

function LogThis(){ # Parameters: LogMessage[text], Abort[0/1], LogDestination[toDNS], LogSeverity[-1 .. -5]
	if [[ $LogToDNS -eq 1 ]] && [[ $3 = "toDNS" ]] && [[ $1 != "" ]]; then
		if [ "$(printf "2.2.1-3051\n$DnsVersion" | sort -V | head -n1)" == "$DnsVersion" ] && [ "$DnsVersion" != "2.2.1-3051" ]; then 
			#Log to file up to 2.2.1-3051
			DTStamp=$(date +"%d-%b-%Y %T.%3N")
			echo "${DTStamp} AdBlock: $1" >> ${LogFile}
		else
			#Log to DB above to 2.2.1-3051
			DTStamp=$(date +"%Y/%m/%d %T.%3N") && LogMsg=$1 && LogMsg=${LogMsg//\'/\'\'}
			# $4=Severity: Info=-1, Notice=-2 ,Warning=-3, Error=-4, Critical=-5
			sqlite3 $LogDB "INSERT INTO LOGS (TIME,TAG,CATAGORY,MODULE,SEVERITY,CONTENT) VALUES ('${DTStamp}','','adblock','ad-blocker.sh','$4','AdBlocker: ${LogMsg}');"
		fi
	fi
	DTStamp=$(date +"%Y-%m-%d %T")
	if [[ $1 = "" ]]; then
		while IFS= read -r line; do
			echo "--${DTStamp}--  $line" >> ${ScriptLog}
		done
	else
		echo "--${DTStamp}--  ad-blocker.sh: $1" >> ${ScriptLog}
		if [ $2 -eq 1 ]; then
			ExitScript 1
		fi
	fi
}
function ExitScript(){ # Parameters: Abort[0/1]
	LogToDNS=1
	if [ $1 -eq 0 ]; then
	    if [ "$Debug" -eq 0 ]; then
            rm -rfv ${TmpDir} | LogThis
            # rm -fv ${ScriptLog}
		fi
		LogThis "Update completed successfully SN ${ZoneSN}" 0 "toDNS" -2
		echo "Update completed successfully SN ${ZoneSN}"
	else
		LogThis "Update aborted with errors, see update log for details ${ScriptLog}" 0 "toDNS" -4
		echo "Update aborted with errors, see update log for details ${ScriptLog}"
	fi
	exit $1
}
# start with an empty log
rm -fv ${ScriptLog} | LogThis

# Check if needed dependencies exists
MissingDep=0
	for NeededDep in 7z cat chown cut date grep mv rm sed sort uniq wget; do
		if ! hash "$NeededDep" >/dev/null 2>&1; then
			LogThis "Command not found in PATH: ${NeededDep}" 0
			MissingDep=$((MissingDep+1))
		fi
	done
	if [ $MissingDep -gt 0 ]; then
		LogThis "Minimum ${MissingDep} commands are missing in PATH, aborting" 1
	fi
	if [ -f $PackageDir/enabled ]; then
    	DnsVersion="$(grep -oP '(?<=version\=\").*(?=\")' ${PackageDir}/INFO)"
	else
		LogThis "DNS Server package not installed, aborting" 1
	fi

#Read user config 
	LogToDNS=1 ; LogThis "Update started by user: $SUDO_USER $(whoami) ..." 0 "toDNS" -2
	#Config sets LogToDNS flag till the exit script function
	if [ -f $ScriptCfg ]; then
		LogThis "Reading user config...." 0
		# check if the file contains something we don't want
		if egrep -q -iv '(^\s*#|^\s*$|^\s*[a-z_][^[:space:]]*=[^;&]*$)' "$ScriptCfg"; then
			LogThis "Config file is unclean, please clean it!" 1
		fi
		sed -r -n -e '/^#|^\n/d' -e '/^[[:alpha:]]([[:print:]]|[[:blank:]])+/p' $ScriptCfg | LogThis
		source "$ScriptCfg"
	else
		LogThis "Missing configuration file ${ScriptCfg}, aborting" 1
	fi
	
#Download blacklists and build raw files to temp folder
	mkdir -pv ${TmpDir} | LogThis
	cd ${TmpDir}
	if [ $? -ne 0 ]; then
		LogThis "Can't change directory to ${TmpDir}" 1
	fi
	rm -fv adbListType*.* | LogThis
	rm -fv adbSourceURLs | LogThis
	if [ -f "${USER_BL}" ]; then
		sed -r -e 's/[[:blank:]]*[\#\/]+[[:print:]]+//g' ${USER_BL} > adbListType3.raw
	fi
	for (( BL_Index=1; BL_Index<=$BL_Items; BL_Index++ )); do if [ ${BL_Enabled[${BL_Index}]} -gt 0 ]; then
		LogThis "Downloading ${BL_Name[${BL_Index}]} ..." 0 "toDNS" -2
		echo "Source[${BL_Index}]: ${BL_URL[${BL_Index}]}" >> adbSourceURLs
		wget -a "${ScriptLog}" -O adbList${BL_Index}.dnl "${BL_URL[${BL_Index}]}"
        if [ $? -ne 0 ]; then
            LogToDNS=1
            LogThis "Error on downloading ${BL_Name[${BL_Index}]}" 0 "toDNS" -2
        else
            # Unzip if necessary
            if [ ${BL_Zipped[${BL_Index}]} -gt 0 ]; then
                LogThis "Unzipping ${BL_Name[${BL_Index}]} into folder '${BL_Index}' ..." 0
                7z e -o${BL_Index} adbList${BL_Index}.dnl | LogThis
                mv -fv ${BL_Index}/* adbList${BL_Index}.raw | LogThis
            else
                mv -fv "${TmpDir}/adbList${BL_Index}.dnl" "${TmpDir}/adbList${BL_Index}.raw" | LogThis
            fi
            LogThis "Linking by tipology '${BL_Name[${BL_Index}]}' to 'adbListType${BL_Type[${BL_Index}]}.raw' ..." 0
            echo "//===== ${BL_Name[${BL_Index}]} ====================================================================" | cut -c1-62 >> "adbListType${BL_Type[${BL_Index}]}.raw"
            cat "${TmpDir}/adbList${BL_Index}.raw" >> "adbListType${BL_Type[${BL_Index}]}.raw"
        fi
	fi done



#Downloaded file conversion - Place to add additional file formats (everything will follow)
	LogThis "Cleaning downloaded files ..." 0 "toDNS" -2
	rm -fv adbListAll.raw | LogThis
	for BL_Index in {0..3}; do if [ -f adbListType$BL_Index.raw ]; then
		#--------------------------------------------------------------------------------------
				case $BL_Index in ### Additional download file format decoders can be added here ###
				[0-1]) LogThis "Cleaning type $BL_Index files ..." 0 # [0-1]standard zone file
					grep -o -E '^[[:blank:]]*zone[[:blank:]]*"([A-Za-z0-9]|_|\-|\~|\.)+\.[A-Za-z]{2,}' adbListType$BL_Index.raw >> adbListAll.raw # Extract domains (2018.02.04)
					;;
				2) LogThis "Cleaning type 2 files ..." 0 # [2]Standard hosts file
					sed -r -n -e '/(localhost|local[[:blank:]]*\n|[0-9]{1,3}\.[0-9]{1,3}\.[0-9]{1,3}\.255|0\.0\.0\.0[[:blank:]]+0\.0\.0\.0)/d' -e '/^(127\.0\.0\.1|0\.0\.0\.0)[[:blank:]]+(\w|\-|\~|\.)+\.[A-Za-z]{2,}/p' adbListType$BL_Index.raw > adbListType$BL_Index.p01 # Clean and Validate (2018.02.04)

					sed -r -e 's/^(127\.0\.0\.1|0\.0\.0\.0)[[:space:]]*/zone "/g' adbListType$BL_Index.p01 >> adbListAll.raw # Add line head (2018.01.07 M4RC)
					;;
				3) LogThis "Cleaning type 3 files ..." 0 # [3]FQDN domain/host(only) list
					sed -r -n -e '/^[[:space:]]*(\w|\-|\~|\.)+\.[A-Za-z]{2,}/p' adbListType$BL_Index.raw > adbListType$BL_Index.p01 # Clean (2018.02.04)
					sed -r -e 's/^[[:space:]]*/zone "/g' adbListType$BL_Index.p01 >> adbListAll.raw # Add line head
					;;
				esac
		#--------------------------------------------------------------------------------------
	fi done
	# Patch uppercase chars and reduces duplicated lines to single line
	sed -e 's/\s*$//g' adbListAll.raw | sed -e 's/\([[:upper:]]\)/\L\1/g' | sort | uniq > adbListAll.all #(2018.01.27 Briolet)



# Whitelist generation and processing
	LogThis "Building combined whitelist ..." 0
	rm -fv adbWhiteList.* | LogThis
	echo "----- DNS Server defined zones ----------------------" > adbWhiteList.raw
	for SynoZone in "$ZoneDataDir"/* ;do
		sed -r -e 's/(^.*\/)|\([0-9]*\)$//g' <<< $SynoZone >> adbWhiteList.p01
	done
	sed -e 's/\([[:upper:]]\)/\L\1/g' adbWhiteList.p01 | sort | uniq >> adbWhiteList.raw # Cleaning and sorting (2018.01.27 Briolet)
	if [ -f "${USER_WL}" ]; then
		echo "----- USER Whitelist entries ------------------------" >> adbWhiteList.raw
		sed -r -e 's/[[:blank:]]*[\#\/]+[[:print:]]+//g' ${USER_WL} | sed -e 's/\([[:upper:]]\)/\L\1/g' | sort | uniq -u >> adbWhiteList.raw #Cleaning and sorting (2018.01.27 Briolet)
	fi
	sed -r -n -e '/ad-blocker.db|null.zone.file/{d;n;}' -e '/^[[:space:]]*([A-Za-z0-9]|-|_|\.)*\.([A-Za-z])*/p' adbWhiteList.raw > adbWhiteList.p01 # Clean and Validate
	#The following lines are added twice to prevent "whitelist" becoming "blacklist" in case off missing downloads/correspondences
	sed -r -e 's/^[[:space:]]*/zone "/g' adbWhiteList.p01 >> adbListAll.all # Add line head (once)
	sed -r -e 's/^[[:space:]]*/zone "/g' adbWhiteList.p01 >> adbListAll.all # Add line head (twice) 

#Database file generation
	LogThis "Building final zones database ..." 0 "toDNS" -2
	echo "=====================================================" > ad-blocker.raw
	echo "===== Ad-Blocker.db =================================" >> ad-blocker.raw
	echo "=====================================================" >> ad-blocker.raw
	echo "Generated automatically by Ad-Blocker.sh V2.03.0118" >> ad-blocker.raw
	echo "Build timestamp: ${DTStamp}" >> ad-blocker.raw
	echo >> ad-blocker.raw
	echo "Together with the 'null.zone.file' records this DB" >> ad-blocker.raw
	echo "will return 127.0.0.1 for the listed domains/hosts" >> ad-blocker.raw
	echo >> ad-blocker.raw
	cat adbWhiteList.raw >> ad-blocker.raw
	echo "----- USER Blacklist entries ------------------------" >> ad-blocker.raw
	# sed -r -n -e '/^[[:space:]]|^\#+|^\/{2,}/d' -e 's/[[:blank:]]*[\#\/]+[[:print:]]+//gp' ${USER_BL} >> ad-blocker.raw # (2018.02.04) (Briolet: this line generates nothing)
	cat "${USER_BL}" >> ad-blocker.raw # (Briolet: this line replaces the one above)
	echo "----- Used source URLs ------------------------------" >> ad-blocker.raw
	cat adbSourceURLs >> ad-blocker.raw
	echo "=====================================================" >> ad-blocker.raw
	sed -e 's/^/\/\//g' ad-blocker.raw > ad-blocker.db
	echo >> ad-blocker.db
	# Remove duplicate pairs whitelistening the DB and add the "suitable line tails" (mainly file path for Synology bind  "/etc/zone/master/null.zone.file")
	sort adbListAll.all | uniq -u | sed -e 's/\.*[[:space:]]*$/" { type master; notify no; file "\/etc\/zone\/master\/null.zone.file"; allow-transfer {none;}; allow-update {none;}; allow-query {any;}; };/g' >> ad-blocker.db
	# Place the DB to the right place
	chown -v DNSServer:DNSServer ad-blocker.db | LogThis
	if [ $? -eq 0 ] && [ -f ad-blocker.db ]; then
		chmod -v 0644 ad-blocker.db | LogThis
		mv -fv ad-blocker.db ${ZoneDataDir}/ad-blocker.db | LogThis
		cp -fv ${ZoneDataDir}/ad-blocker.db ${ScriptDir}/final-list.txt | LogThis # Copy list to script dir for user reading
	else 
		LogThis "Script requires root permissions" 1
	fi

# Include the new zone data
	LogThis "Updating DNS Server files ..." 0
	cd ${ZoneDataDir}
	if [ $? -ne 0 ]; then
		LogThis "Can't change directory to ${ZoneDataDir}" 1
	fi
	if [ -f ad-blocker.db ] && [ -f null.zone.file ]; then
		# Rebuild data null.zone.file
		echo 'zone "null.zone.file" {'					> null.zone.file
		echo '	type master;'							>> null.zone.file
		echo '	file "/etc/zone/master/null.zone.file";'>> null.zone.file
		echo '	allow-transfer {any;};'					>> null.zone.file
		echo '	allow-update {none;};'					>> null.zone.file
		echo '	allow-query {any;};'					>> null.zone.file
		echo '};'										>> null.zone.file
		echo 'include "/etc/zone/data/ad-blocker.db";'	>> null.zone.file
		chown -v DNSServer:DNSServer null.zone.file | LogThis
		chmod -v 0644 null.zone.file | LogThis

		# Rebuild master null.zone.file
		cd ${ZoneMasterDir}
		if [ $? -ne 0 ]; then
			LogThis "Can't change directory to ${ZoneMasterDir}" 1
		fi
		CurrSN=$(sed -nr '/@[[:blank:]+]IN[[:blank:]+]SOA|null\.zone\.file\.[[:blank:]+]IN[[:blank:]+]SOA/{/[0-9]/!N;s/[^0-9]+([0-9]+).*/\1/p}' 'null.zone.file')
		Now=$(date +%Y%m%d)
		if [[ "${Now}" -ne "${CurrSN::-2}" || ${#CurrSN} -lt ${#Now} ]]; then
			ZoneSN="${Now}00" #Set new timestamp
		else
			ZoneSN="${Now}$(printf '%02d' $((10#${CurrSN: -2} + 1)) )" # Increment if already same timestamp
		fi
		echo '$TTL 86400				; one day'				> null.zone.file
		echo '@ IN	SOA	ns.null.zone.file. mail.null.zone.file. (' >> null.zone.file
		echo '			'${ZoneSN}'	; serial number YYYYMMDDNN'	>> null.zone.file
		echo '			86400		; refresh	1 day'			>> null.zone.file
		echo '			7200		; retry		2 hours'		>> null.zone.file
		echo '			864000		; expire	10 days'		>> null.zone.file
		echo '			86400 )		; min ttl	1 day'			>> null.zone.file
		echo '		NS	ns.null.zone.file.'						>> null.zone.file
		echo '		A	127.0.0.1'								>> null.zone.file
		echo '		AAAA	::1'								>> null.zone.file
		echo '* IN	A	127.0.0.1'								>> null.zone.file
		echo '* IN	AAAA	::1'								>> null.zone.file
		chown -v DNSServer:DNSServer null.zone.file | LogThis
		chmod -v 0644 null.zone.file | LogThis
		LogThis "null.zone.file SN changed from ${CurrSN} to ${ZoneSN}" 0
	else
		LogThis "File not found, check for presence of null.zone.file domain zone" 1
	fi

# Reload the server config after modifications
	if ${PackageDir}/scripts/start-stop-status status; then
		LogThis "Restarting DNS Server ..." 0
		${RootDir}/script/reload.sh | LogThis
	else
		LogThis "DNS Server not started ..." 0
	fi

ExitScript 0